;
; rspells.inc - rythm spells
;
;
; Copyright (C) 2006  Nathan (Acorn) Pooley
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; version 2 as published by the Free Software Foundation.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License (gpl.txt) for more details. 
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
; 
; You can contact the author, Nathan (Acorn) Pooley, by writing
; to Nathan (Acorn) Pooley, 949 Buckeye Drive, Sunnyvale, CA  94086, USA
; or through the form at http://www.rawbw.com/~acorn/wand/feedback.html
; 
;#@DOC@ definitions of rythm spells (song spells)


;
; special bytes
;
; SPELL_END  (0x00)  - terminate spell - ignored elsewhere
; SPELL_LAST         - follows last spell
; SPELL_PARTIAL      - begin partial spells
;
; Spell info h,l follows after terminating 0x00
;


;
; length of longest and shortest rythms
;
RSPELL_MIN_TAPS		equ	5	; notes in shortest song
RSPELL_MAX_TAPS		equ	26	; notes in longest song
RSPELL_LATEST_ONE	equ	10	; latest "1" in a song


;
; These are terminated (complete songs)
;

w_rspell1:
w_rspell_shave1:	; shave and a haircut
    db      0x21, 0x12, 0x42, 0x00
	db		HIGH w_spellinfo_shave, LOW w_spellinfo_shave

w_rspell_shave2:	; shave and a haircut
    db      0x32, 0x13, 0x63, 0x00
	db		HIGH w_spellinfo_shave, LOW w_spellinfo_shave

w_rspell_row:		; row row row your boat
    db      0x33, 0x21, 0x32, 0x12, 0x16, 0x11, 0x11, 0x11 
    db      0x11, 0x11, 0x11, 0x21, 0x21, 0x00
	db		HIGH w_spellinfo_row, LOW w_spellinfo_row

w_rspell_jeff1:		; jeff's made up little song
    db      0x22, 0x11, 0x11, 0x22
	db		0x00, HIGH w_spellinfo_jeff1, LOW w_spellinfo_jeff1

#if 0
w_rspell_anthem1:	; my country tis of thee
	db		0x44, 0x46, 0x24, 0x44, 0x46, 0x24, 0x44, 0x48
	db		0x44, 0x46, 0x24, 0x44, 0x46, 0x24, 0x42, 0x22
	db		0x26, 0x24, 0x47, 0x10
	db		0x00, HIGH w_spellinfo_anthem, LOW w_spellinfo_anthem
#endif

w_rspell_quarters:	; westminster quarters
	db		0x11, 0x13, 0x11, 0x13, 0x11, 0x13, 0x11, 0x10
	db		0x00, HIGH w_spellinfo_quarters, LOW w_spellinfo_quarters


	db		SPELL_PARTIAL


;
; These are run-on (partial songs)
;
w_rspell_twinkle:	; twinkle twinkle little star
    db      0x11, 0x11, 0x11, 0x21, 0x11, 0x11
	db		0x00, HIGH w_spellinfo_twinkle, LOW w_spellinfo_twinkle

w_rspell_puffthemagicdragon:
	db		0x31, 0x22, 0x44, 0x42
	db		0x00, HIGH w_spellinfo_puffthemagicdragon, LOW w_spellinfo_puffthemagicdragon

w_rspell_jingle:	; jingle bells
    db      0x22, 0x42, 0x24, 0x22, 0x31, 0x82, 0x23, 0x12
    db      0x22, 0x11, 0x22, 0x22
	db		0x00, HIGH w_spellinfo_jingle, LOW w_spellinfo_jingle

w_rspell_mary:		; mary had a little lamb
    db      0x11, 0x11, 0x11, 0x21, 0x12, 0x11, 0x21, 0x11
    db      0x11, 0x11, 0x11, 0x11
	db		0x00, HIGH w_spellinfo_mary, LOW w_spellinfo_mary

w_rspell_brady:		; the brady bunch theme
    db      0x22, 0x28, 0x11, 0x22, 0x28, 0x22, 0x22, 0x00
	db		HIGH w_spellinfo_brady, LOW w_spellinfo_brady

w_rspell_gilligan:	; gilligans island theme
    db      0x42, 0x22, 0x11, 0x22, 0x31, 0x21, 0x12, 0x00
	db		HIGH w_spellinfo_gilligan, LOW w_spellinfo_gilligan

w_rspell_flint1:		; the flintstones theme
    db      0x23, 0x21, 0x23, 0x21, 0x11, 0x11, 0x22, 0x00
	db		HIGH w_spellinfo_flint, LOW w_spellinfo_flint

w_rspell_flint2:		; the flintstones theme
    db      0x24, 0x21, 0x24, 0x21, 0x11, 0x11, 0x22, 0x00
	db		HIGH w_spellinfo_flint, LOW w_spellinfo_flint

w_rspell_star:		; star spangled banner (BROKEN)
    db      0x22, 0x44, 0x48, 0x31, 0x44, 0x48, 0x22, 0x62
	db		0x48, 0x22
	db		0x00, HIGH w_spellinfo_star, LOW w_spellinfo_star

w_rspell_birthday:	; happy birthday to you
    db      0x21, 0x33, 0x36, 0x21, 0x33, 0x00
	db		HIGH w_spellinfo_birthday, LOW w_spellinfo_birthday

#if 0
w_rspell_anthem2:	; my country tis of thee
	db		0x22, 0x23, 0x12, 0x22, 0x23, 0x12, 0x22, 0x24
	db		0x22, 0x23, 0x12, 0x22, 0x23, 0x12, 0x21, 0x11
	db		0x13, 0x12
	db		0x00, HIGH w_spellinfo_anthem, LOW w_spellinfo_anthem
#endif

w_rspell_godsavethequeen:	; my country tis of thee
	db		0x22, 0x23, 0x12, 0x22, 0x23, 0x12, 0x22, 0x00
	db		HIGH w_spellinfo_godsavethequeen, LOW w_spellinfo_godsavethequeen



	db		SPELL_LAST


